#!/usr/bin/env python
# encoding: utf-8
#
# ** header v3.0
# This file is a part of the CaosDB Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
# Copyright (C) 2019 Henrik tom Wörden
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
""" Defines how something that shall be inserted into CaosDB is treated.

CaosDB can automatically be filled with Records based on some file structure.
The Crawler will iterate over the files and test for each file whether a CFood
exists that matches the file path. If one does, it is instanciated to treat the
match. This occurs in basically three steps:
1. create a list of identifiables, i.e. unique representation of CaosDB Records
(such as an experiment belonging to a project and a date/time)
2. the identifiables are either found in CaosDB or they are created.
3. the identifiables are update based on the date in the file structure
"""

import logging
import re

import caosdb as db
from caosdb.exceptions import EntityDoesNotExistError

from .guard import global_guard as guard

ENTITIES = {}

logger = logging.getLogger(__name__)


def get_entity(name):
    """ Returns the entity with a given name, preferably from a local cache.

    If the local cache does not contain the entity, retrieve it from CaosDB.
    """

    if name not in ENTITIES:
        ent = db.Entity(name=name)
        ent.retrieve()
        ENTITIES[name] = ent

    return ENTITIES[name]


class AbstractCFood(object):
    # contains the compiled regular expression after the first execution of the
    # function match()
    _pattern = None

    def __init__(self, crawled_file, access=lambda x: x):
        """ Abstract base class for Crawler food (CFood).

        Parameters
        ----------
        crawled_file : The file that the crawler is currently matching. Its
                       path should match against the pattern of this class

        access : callable, optional
                 A function that takes a CaosDB path and returns a local path
        """
        self.access = access
        self._crawled_file = None
        self.crawled_path = crawled_file
        self.match = type(self).match(crawled_file)
        self.to_be_updated = db.Container()
        self.identifiables = db.Container()
        self.attached_ones = []
        self.attached_filenames = []

    @property
    def crawled_file(self):
        if self._crawled_file is None:
            try:
                q = "FIND FILE WHICH IS STORED AT '{}'".format(
                    self.crawled_path)
                self._crawled_file = db.execute_query(q, unique=True)
            except EntityDoesNotExistError:
                path = "**"

                if not self.crawled_path.startswith("/"):
                    path = path + "/"
                q = "FIND FILE WHICH IS STORED AT '{}{}'".format(path,
                                                                 self.crawled_path)
                logger.debug(q)
                self._crawled_file = db.execute_query(q, unique=True)

        return self._crawled_file

    def collect_information(self):
        """ The CFood collects information for further processing.

        Often CFoods need information from files or even from the database in
        order to make processing decision. It is intended that this function is
        called after match. Thus match can be used without connecting to the
        database.

        To be overwritten by subclasses
        """
        pass

    @staticmethod
    def get_re():
        """ Returns the regular expression used to identify files that shall be
        processed

        This function shall be implemented by subclasses.
        """
        raise NotImplementedError()

    @classmethod
    def cook(cls, crawled_file, **kwargs):
        """ possibly checks for existing CFoods whether the match should be
        added or whether a new CFood instance needs to be returned

        This function should typically be used to create CFoods in order to
        prevent the creation of unnecessary instances.

        This standard implementation does not do a check but may be overwritten
        by subclasses.

        Retruns
        -------------
        CFood: if a new instance was created
        None: otherwise
        """

        return cls(crawled_file, **kwargs)

    @classmethod
    def match(cls, string):
        """ Matches the regular expression of this class against file names

        Parameters
        ----------
        string : str
                 The path of the file that shall be matched.
        """

        # TODO this does not quite work. Sometimes the wrong expression is in
        # _pattern; FIX
        # if cls._pattern is None:
        #    cls._pattern = re.compile(cls.get_re())

        # return cls._pattern.match(string)

        return re.match(cls.get_re(), string)

    def create_identifiables(self):
        """
        should set the instance variable Container with the identifiables
        """
        raise NotImplementedError()

    def update_identifiables(self):
        """ Changes the identifiables as needed and adds changed identifiables
        to self.to_be_updated
        """
        raise NotImplementedError()

    def push_identifiables_to_CaosDB(self):
        """ Updates the self.to_be_updated Container, i.e. pushes the changes
        to CaosDB
        """

        if len(self.to_be_updated) == 0:
            return

        get_ids_for_entities_with_names(self.to_be_updated)

        # remove duplicates
        tmp = db.Container()

        for el in self.to_be_updated:
            if el not in tmp:
                tmp.append(el)

        self.to_be_updated = tmp

        logger.info("UPDATE: updating the following entities")

        for el in self.to_be_updated:
            logger.info("\t" + el.name if el.name is not None else el.id)

        logger.debug(self.to_be_updated)
        guard.safe_update(self.to_be_updated)

    def attach(self, crawled_file):
        self.attached_ones.append(crawled_file)

    def looking_for(self, crawled_file):
        """
        returns True if crawled_file can be added to this CFood.

        Typically a CFood exists for a file and defines how to deal with the
        file. However, sometimes additional files "belong" to a CFood. E.g. an
        experiment CFood might match against a README file but labnotes.txt
        also shall be treated by the cfood (and not a special cfood created for
        labnotes.txt)
        This function can be used to define what files shall be 'attached'.
        """

        if crawled_file in self.attached_filenames:
            return True

        return False

    @staticmethod
    # move to api?
    def set_parents(entity, names):
        entity.parents.clear()

        for n in names:
            entity.add_parent(get_entity(n))

    @staticmethod
    # move to api?
    def remove_property(entity, prop):
        # TODO only do something when it is necessary?

        if isinstance(prop, db.Entity):
            name = prop.name
        else:
            name = prop

        while entity.get_property(name) is not None:
            entity.remove_property(name)

    @staticmethod
    # move to api?
    def set_property(entity, prop, value, datatype=None):
        AbstractCFood.remove_property(entity, prop)

        if datatype is not None:
            entity.add_property(prop, value, datatype=datatype)
        else:
            entity.add_property(prop, value)


def assure_object_is_in_list(obj, containing_object, property_name,
                             to_be_updated, datatype=None):
    """
    Checks whether `obj` is one of the values in the list property
    `property_name` of the supplied entity  containing_object`.

    If this is the case this function returns. Otherwise the entity is added to
    the property `property_name` and the entity `containing_object` is added to
    the supplied list to_be_updated in order to indicate, that the entity
    `containing_object` should be updated.

    If the property is missing, it is added first and then the entity is added.

    If obj is a list, every element is added
    """

    if datatype is None:
        datatype = db.LIST(property_name)

    if containing_object.get_property(property_name) is None:
        containing_object.add_property(property_name, value=[],
                                       datatype=datatype)

    if not isinstance(containing_object.get_property(property_name).value, list):
        containing_object.get_property(property_name).value = [containing_object.get_property(property_name).value]
        containing_object.get_property(property_name).value
        containing_object.get_property(property_name).datatype = datatype
    current_list = containing_object.get_property(property_name).value

    if not isinstance(obj, list):
        objects = [obj]
    else:
        objects = obj

    # use ids if values are entities

    if all([isinstance(el, db.Entity) for el in objects]):
        objects = [el.id for el in objects]

    update = False

    for o in objects:
        contained = False

        for el in current_list:
            if el == o:
                contained = True

                break

        if contained:
            logger.debug("{} is in {} of entity {}".format(
                    o, property_name, containing_object.id))

        else:
            logger.info("UPDATE: Appending {} to {} of entity {}".format(
                    o, property_name, containing_object.id))
            current_list.append(o)
            update = True

    if update:
        to_be_updated.append(containing_object)


def assure_has_description(entity, description, to_be_updated=None,
                           force=False):
    """
    Checks whether `entity` has the description that is passed.

    If this is the case this function ends. Otherwise the entity is assigned
    a new description. The list to_be_updated is supplied, the entity is added to
    the list in order to indicate, that the entity `entity` should be updated.
    Otherwise it is directly updated
    """

    if entity.description == description:
        return

    logger.info("UPDATE: set description of entity {}".format(entity.id))
    entity.description = description

    if to_be_updated is None:
        if force:
            entity.update(unique=False)
        else:
            guard.safe_update(entity, unique=False)
    else:
        to_be_updated.append(entity)


def assure_has_parent(entity, parent, to_be_updated=None,
                      force=False, unique=True):
    """
    Checks whether `entity` has a parent with name `parent`.

    If this is the case this function ends. Otherwise the entity is assigned
    a new parent. The list to_be_updated is supplied, the entity is added to
    the list in order to indicate, that the entity `entity` should be updated.
    Otherwise it is directly updated
    """
    parents = entity.get_parents()
    contained = False

    for el in parents:
        if el.name == parent:
            contained = True

            break

    if contained:
        logger.debug("entity {} has parent {}".format(entity.id, parent))

        return

    logger.info("UPDATE: Adding  parent {} to entity {}".format(parent,
                                                                entity.id))
    entity.add_parent(parent)

    if to_be_updated is None:
        get_ids_for_entities_with_names([entity])

        # TODO move the unique argument?
        # TODO  find a better way then force?

        if force:
            entity.update(unique=unique)
        else:
            guard.safe_update(entity, unique=unique)
    else:
        to_be_updated.append(entity)


def assure_has_property(entity, name, value, to_be_updated=None,
                        datatype=None):
    """
    Checks whether `entity` has a property `name` with the value `value`.

    If this is the case this function ends. Otherwise the entity is assigned
    a new parent. The list to_be_updated is supplied, the entity is added to
    the list in order to indicate, that the entity `entity` should be updated.
    Otherwise it is directly updated
    """

    if name.lower() == "description":
        if entity.description == value:
            return
        else:
            logger.info("UPDATE: Adding  description with value {} to "
                        "entity {}".format(value, entity.id))
            entity.description = value

            if to_be_updated is None:
                get_ids_for_entities_with_names([entity])

                guard.safe_update(entity, unique=False)

            return

    properties = entity.get_properties()
    possible_properties = [prop for prop in properties if prop.name.lower() ==
                           name.lower()]
    contained = False

    for el in possible_properties:
        if el.value == value:
            contained = True

            break

    if contained:
        logger.debug("entity {} has property  {} with value {}".format(
                entity.id, name, value))

        return

    logger.info(
        "UPDATE: Adding  property {} with value {} to entity {}".format(
            name, value, entity.id))

    if datatype is None:
        entity.add_property(name=name, value=value)
    else:
        entity.add_property(name=name, value=value, datatype=datatype)

    if to_be_updated is None:
        get_ids_for_entities_with_names([entity])

        guard.safe_update(entity, unique=False)
    else:
        to_be_updated.append(entity)


def insert_id_based_on_name(entity):
    if entity.name is not None and (entity.id is None or entity.id < 0):
        entity.id = get_entity(entity.name).id


def get_ids_for_entities_with_names(entities):
    for ent in entities:
        insert_id_based_on_name(ent)

        for prop in ent.get_properties():
            insert_id_based_on_name(prop)

        for parent in ent.get_parents():
            insert_id_based_on_name(parent)
            insert_id_based_on_name(ent)


class CMeal(object):
    """
    CMeal groups equivalent CFoods and allow their collected insertion.

    Sometimes there is no one file that can be used to trigger the creation of
    some Record. E.g. if a collection of images shall be referenced from one
    Record that groups them, it is unclear which image should trigger the
    creation of the Record.

    CMeals are grouped based on the groups in the used regular expression. If,
    in the above example, all the images reside in one folder, all groups
    except that for the file name should match. The groups that shall match
    need to be listed in the matching_groups class property. Subclasses will
    overwrite this property.

    The cook function of a cfood allows this class to work. Instead of directly
    instantiating a CFood the cook function is used. If the CFood is also a
    child of CMeal, it will be checked (using get_suitable_cfood) in the cook
    function whether a new CFood should be created or if the file match should
    be added to an existing one. In order to allow this all instances of a
    CFood class are tracked in the existing_instances class member.
    """
    existing_instances = []
    matching_groups = []

    def __init__(self, *args, **kwargs):
        self.existing_instances.append(self)
        self.crawled_files = []

    def add(self, crawled_file):
        self.crawled_files.append(crawled_file)

    @classmethod
    def get_suitable_cfood(cls, match):
        for cfood in cls.existing_instances:
            suitable = True

            for group in cls.matching_groups:
                if (group not in match.groupdict() or
                        group not in cfood.match.groupdict() or
                        match.group(group) != cfood.match.group(group)):
                    suitable = False

            if suitable:
                return cfood

        return None
